package net.minecraftforge.common;

import java.util.*;

import cpw.mods.fml.common.FMLLog;
import net.minecraft.world.biome.*;
import net.minecraftforge.event.terraingen.DeferredBiomeDecorator;
import static net.minecraft.world.biome.BiomeGenBase.*;
import static net.minecraftforge.common.BiomeDictionary.Type.*;

public class BiomeDictionary
{
    public enum Type
    {
        FOREST,
        PLAINS,
        MOUNTAIN,
        HILLS,
        SWAMP,
        WATER,
        DESERT,
        FROZEN,
        JUNGLE,
        WASTELAND,
        BEACH,
        NETHER,
        END,
        MUSHROOM,
        MAGICAL;
    }

    private static final int BIOME_LIST_SIZE = BiomeGenBase.func_150565_n().length;
    private static BiomeInfo[] biomeList = new BiomeInfo[BIOME_LIST_SIZE];
    @SuppressWarnings("unchecked")
    private static ArrayList<BiomeGenBase>[] typeInfoList = new ArrayList[Type.values().length];

    private static class BiomeInfo
    {
        public EnumSet<Type> typeList;

        public BiomeInfo(Type[] types)
        {
            typeList = EnumSet.noneOf(Type.class);
            for(Type t : types)
            {
                typeList.add(t);
            }
        }
    }

    static
    {
        registerVanillaBiomes();
    }

    /**
     * Registers a biome with a specific biome type
     * 
     * @param biome the biome to be registered
     * @param type the type to register the biome as
     * @return returns true if the biome was registered successfully
     */
    public static boolean registerBiomeType(BiomeGenBase biome, Type ... types)
    {   
        if(BiomeGenBase.func_150565_n()[biome.field_76756_M] != null)
        {
            for(Type type : types)
            {
                if(typeInfoList[type.ordinal()] == null)
                {
                    typeInfoList[type.ordinal()] = new ArrayList<BiomeGenBase>();
                }

                typeInfoList[type.ordinal()].add(biome);
            }

            if(biomeList[biome.field_76756_M] == null)
            {
                biomeList[biome.field_76756_M] = new BiomeInfo(types);
            }
            else
            {
                for(Type type : types)
                {
                    biomeList[biome.field_76756_M].typeList.add(type);
                }
            }

            return true;
        }

        return false;
    }

    /**
     * Returns a list of biomes registered with a specific type
     * 
     * @param type the Type to look for
     * @return a list of biomes of the specified type, null if there are none
     */
    public static BiomeGenBase[] getBiomesForType(Type type)
    {
        if(typeInfoList[type.ordinal()] != null)
        {
            return (BiomeGenBase[])typeInfoList[type.ordinal()].toArray(new BiomeGenBase[0]);
        }

        return new BiomeGenBase[0];
    }

    /**
     * Gets a list of Types that a specific biome is registered with
     * 
     * @param biome the biome to check
     * @return the list of types, null if there are none
     */
    public static Type[] getTypesForBiome(BiomeGenBase biome)
    {
        checkRegistration(biome);

        if(biomeList[biome.field_76756_M] != null)
        {
            return (Type[])biomeList[biome.field_76756_M].typeList.toArray(new Type[0]);
        }

        return new Type[0];
    }

    /**
     * Checks to see if two biomes are registered as having the same type
     * 
     * @param biomeA
     * @param biomeB
     * @return returns true if a common type is found, false otherwise
     */
    public static boolean areBiomesEquivalent(BiomeGenBase biomeA, BiomeGenBase biomeB)
    {
        int a = biomeA.field_76756_M;
        int b = biomeB.field_76756_M;

        checkRegistration(biomeA);
        checkRegistration(biomeB);

        if(biomeList[a] != null && biomeList[b] != null)
        {
            for(Type type : biomeList[a].typeList)
            {
                if(containsType(biomeList[b], type))
                {
                    return true;
                }
            }
        }

        return false;
    }

    /**
     * Checks to see if the given biome is registered as being a specific type
     * 
     * @param biome the biome to be considered
     * @param type the type to check for
     * @return returns true if the biome is registered as being of type type, false otherwise
     */
    public static boolean isBiomeOfType(BiomeGenBase biome, Type type)
    {
        checkRegistration(biome);

        if(biomeList[biome.field_76756_M] != null)
        {
            return containsType(biomeList[biome.field_76756_M], type);
        }

        return false;
    }

    /**
     * Checks to see if the given biome has been registered as being of any type
     * @param biome the biome to consider
     * @return returns true if the biome has been registered, false otherwise
     */
    public static boolean isBiomeRegistered(BiomeGenBase biome)
    {    
        return biomeList[biome.field_76756_M] != null;
    }

    public static boolean isBiomeRegistered(int biomeID)
    {
        return biomeList[biomeID] != null;
    }

    public static void registerAllBiomes()
    {
        FMLLog.warning("Redundant call to BiomeDictionary.registerAllBiomes ignored");
    }
    /**
     * Loops through the biome list and automatically adds tags to any biome that does not have any
     * This is called by Forge at postinit time. It will additionally dispatch any deferred decorator
     * creation events.
     * 
     * DO NOT call this during world generation
     */
    public static void registerAllBiomesAndGenerateEvents()
    {
        for(int i = 0; i < BiomeGenBase.func_150565_n().length; i++)
        {
            BiomeGenBase biome = BiomeGenBase.func_150565_n()[i];

            if(biome == null)
            {
                continue;
            }

            if (biome.field_76760_I instanceof DeferredBiomeDecorator)
            {
                DeferredBiomeDecorator decorator = (DeferredBiomeDecorator) biome.field_76760_I;
                decorator.fireCreateEventAndReplace(biome);
            }

            checkRegistration(biome);
        }
    }

    /**
     * Automatically looks for and registers a given biome with appropriate tags
     * This method is called automatically if a biome has not been registered with any tags,
     * And another method requests information about it
     * 
     * NOTE: You can opt out of having your biome registered by registering it as type NULL
     * 
     * @param biome the biome to be considered
     */
    public static void makeBestGuess(BiomeGenBase biome)
    {    
        if(biome.field_76760_I.field_76832_z >= 3)
        {
            if(biome.func_76736_e() && biome.field_76750_F >= 1.0F)
            {
                BiomeDictionary.registerBiomeType(biome, JUNGLE);
            }
            else if(!biome.func_76736_e())
            {
                BiomeDictionary.registerBiomeType(biome, FOREST);
            }
        }
        else if(biome.field_76749_E <= 0.3F && biome.field_76749_E >= 0.0F)
        {
            if(!biome.func_76736_e() || biome.field_76748_D >= 0.0F)
            {
                BiomeDictionary.registerBiomeType(biome, PLAINS);
            }
        }

        if(biome.func_76736_e() && biome.field_76748_D < 0.0F && (biome.field_76749_E <= 0.3F && biome.field_76749_E >= 0.0F))
        {
            BiomeDictionary.registerBiomeType(biome, SWAMP);
        }

        if(biome.field_76748_D <= -0.5F)
        {
            BiomeDictionary.registerBiomeType(biome, WATER);
        }

        if(biome.field_76749_E >= 1.5F)
        {
            BiomeDictionary.registerBiomeType(biome, MOUNTAIN);
        }
        
        if(biome.func_76746_c() || biome.field_76750_F < 0.2F)
        {
            BiomeDictionary.registerBiomeType(biome, FROZEN);
        }
        
        if(!biome.func_76736_e() && biome.field_76750_F >= 1.0F)
        {
            BiomeDictionary.registerBiomeType(biome, DESERT);
        }
    }

    //Internal implementation    
    private static void checkRegistration(BiomeGenBase biome)
    {
        if(!isBiomeRegistered(biome))
        {
            makeBestGuess(biome);
        }
    }

    private static boolean containsType(BiomeInfo info, Type type)
    {
        return info.typeList.contains(type);
    }

    private static void registerVanillaBiomes()
    {
        registerBiomeType(field_76771_b,               WATER          );
        registerBiomeType(field_76772_c,              PLAINS         );
        registerBiomeType(field_76769_d,              DESERT         );
        registerBiomeType(field_76770_e,        MOUNTAIN       );
        registerBiomeType(field_76767_f,              FOREST         );
        registerBiomeType(field_76768_g,               FOREST,  FROZEN);
        registerBiomeType(field_76784_u,          FOREST,  FROZEN);
        registerBiomeType(field_76780_h,           SWAMP          );
        registerBiomeType(field_76781_i,               WATER          );
        registerBiomeType(field_76776_l,         WATER,   FROZEN);
        registerBiomeType(field_76777_m,         WATER,   FROZEN);
        registerBiomeType(field_76774_n,           FROZEN         );
        registerBiomeType(field_76775_o,        FROZEN         );
        registerBiomeType(field_76787_r,               BEACH          );
        registerBiomeType(field_76786_s,         DESERT         );
        registerBiomeType(field_76782_w,              JUNGLE         );
        registerBiomeType(field_76792_x,         JUNGLE         );
        registerBiomeType(field_76785_t,         FOREST         );
        registerBiomeType(field_76779_k,                 END            );
        registerBiomeType(field_76778_j,                NETHER         );
        registerBiomeType(field_76789_p,      MUSHROOM       );
        registerBiomeType(field_76783_v,    MOUNTAIN       );
        registerBiomeType(field_76788_q, MUSHROOM, BEACH);
    }
}
